#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Transliteration module of the Kurdish Language Processing Toolkit (KLPT).

    Based on the Wergor project: https://github.com/sinaahmadi/wergor
    Ahmadi, Sina. "A rule-based Kurdish text transliteration system." ACM Transactions on Asian and Low-Resource Language Information Processing (TALLIP) 18.2 (2019): 1-8.
    
    * created: 2020/06/12 02:13:26
    * author: Sina Ahmadi

"""

import sys
# sys.path.append('../klpt')
import itertools
import json
import re
from .preprocess import Preprocess
from .configuration import Configuration
import klpt

class Transliterate:
    """
    This module aims at transliterating one script of Kurdish into another one. Currently, only the Latin-based and the Arabic-based scripts of Sorani and Kurmanji are supported. The main function in this module is `transliterate()` which also takes care of detecting the correct form of double-usage graphemes, namely و ↔ w/u and ی ↔ î/y. In some specific occasions, it can also predict the placement of the missing *i* (also known as *Bizroke/بزرۆکە*).

    The module is based on the [Kurdish transliteration project](https://github.com/sinaahmadi/wergor).

    Example:
    ```python
    >>> from klpt.transliterate import Transliterate
    >>> transliterate = Transliterate("Kurmanji", "Latin", target_script="Arabic")
    >>> transliterate.transliterate("rojhilata navîn")
    'رۆژهلاتا ناڤین'

    >>> transliterate_ckb = Transliterate("Sorani", "Arabic", target_script="Latin")
    >>> transliterate_ckb.transliterate("لە وڵاتەکانی دیکەدا")
    'le wiłatekanî dîkeda'
    ```

    """

    def __init__(self, dialect, script, target_script, unknown="�", numeral="Latin"):
        """Initializing using a Configuration object

        To do:
            - "لە ئیسپانیا ژنان لە دژی ‘patriarkavirus’ ڕێپێوانیان کرد": "le îspanya jinan le dijî ‘patriarkavirus’ řêpêwanyan kird"
            - "egerçî damezrandnî rêkxrawe kurdîyekan her rêpênedraw mabûnewe Inzîbat.": "ئەگەرچی دامەزراندنی ڕێکخراوە کوردییەکان هەر رێپێنەدراو مابوونەوە ئنزیبات.",

        Args:
            mode ([type]): [description]
            unknown (str, optional): [description]. Defaults to "�".
            numeral (str, optional): [description]. Defaults to "Latin". Modifiable only if the source script is in Arabic. Otherwise, the Default value will be Latin.

        Raises:
            ValueError: [description]
            ValueError: [description]

        """
        # with open("data/default-options.json") as f:
        #     options = json.load(f)
        
        self.UNKNOWN = "�"
        with open(klpt.get_data("data/wergor.json"), encoding = "utf-8") as f:
            self.wergor_configurations = json.load(f)

        with open(klpt.get_data("data/preprocess_map.json"), encoding = "utf-8") as f:
            self.preprocess_map = json.load(f)["normalizer"]
        
        configuration = Configuration({"dialect": dialect, "script": script, "numeral": numeral, "target_script": target_script, "unknown": unknown})
        # self.preprocess_map = object.preprocess_map["normalizer"]
        self.dialect = configuration.dialect
        self.script = configuration.script
        self.numeral = configuration.numeral
        self.mode = configuration.mode
        self.target_script = configuration.target_script
        self.user_UNKNOWN = configuration.user_UNKNOWN

        # self.mode = mode
        # if mode=="arabic_to_latin":
        #     target_script = "Latin"
        # elif mode=="latin_to_arabic":
        #     target_script = "Arabic"
        # else:
        #     raise ValueError(f'Unknown transliteration option. Available options: {options["transliterator"]}')
    
        # if len(unknown):
        #     self.user_UNKNOWN = unknown
        # else:
        #     raise ValueError(f'Unknown unknown tag. Select a non-empty token (e.g. <UNK>.')

        self.characters_mapping = self.wergor_configurations["characters_mapping"]
        self.digits_mapping = self.preprocess_map["universal"]["numerals"][self.target_script]
        self.digits_mapping_all = list(set(list(self.preprocess_map["universal"]["numerals"][self.target_script].keys()) + list(self.preprocess_map["universal"]["numerals"][self.target_script].values())))
        self.punctuation_mapping = self.wergor_configurations["punctuation"][self.target_script]
        self.punctuation_mapping_all = list(set(list(self.wergor_configurations["punctuation"][self.target_script].keys()) + 
                                                list(self.wergor_configurations["punctuation"][self.target_script].values())))
        # self.tricky_characters = self.wergor_configurations["characters_mapping"]
        self.wy_mappings = self.wergor_configurations["wy_mappings"]

        self.hemze = self.wergor_configurations["hemze"]
        self.bizroke = self.wergor_configurations["bizroke"]
        self.uw_iy_forms = self.wergor_configurations["uw_iy_forms"]
        self.target_char = self.wergor_configurations["target_char"]
        self.arabic_vowels = self.wergor_configurations["arabic_vowels"]
        self.arabic_cons = self.wergor_configurations["arabic_cons"]
        self.latin_vowels = self.wergor_configurations["latin_vowels"]
        self.latin_cons = self.wergor_configurations["latin_cons"]
        
        self.characters_pack = {"arabic_to_latin": self.characters_mapping.values(), "latin_to_arabic": self.characters_mapping.keys()}
        if self.target_script == "Arabic":
            self.prep = Preprocess("Sorani", "Latin", numeral=self.numeral)
        else:
            self.prep = Preprocess("Sorani", "Latin", numeral="Latin")


    def to_pieces(self, token):
        """Given a token, find other segments composed of numbers and punctuation marks not seperated by space ▁""" 
        tokens_dict = dict()
        flag = False # True if a token is a \w
        i = 0

        for char_index in range(len(token)):
            if token[char_index] in self.digits_mapping_all or token[char_index] in self.punctuation_mapping_all:
                tokens_dict[char_index] = token[char_index]
                flag = False
                i = 0
            elif token[char_index] in self.characters_pack[self.mode] or \
                token[char_index] in self.target_char or \
                token[char_index] == self.hemze or token[char_index].lower() == self.bizroke:
                if flag:
                    tokens_dict[char_index-i] = tokens_dict[char_index-i] + token[char_index]
                else:
                    tokens_dict[char_index] = token[char_index]
                flag = True
                i += 1
            else:
                tokens_dict[char_index] = self.UNKNOWN
        
        return tokens_dict

    def transliterate(self, text):
        """The main method of the class:

        - find word boundaries by splitting it using spaces and then retrieve words mixed with other characters (without space)
        - map characters
        - detect double-usage characters w/u and y/î
        - find possible position of Bizroke (to be completed - 2017)

        Notice: text format should not be changed at all (no lower case, no style replacement \t, \n etc.).
        If the source and the target scripts are identical, the input text should be returned without any further processing.

        """
        text = self.prep.unify_numerals(text).split("\n")
        transliterated_text = list()

        for line in text:
            transliterated_line = list()
            for token in line.split():
                trans_token = ""
                # try:
                token = self.preprocessor(token) # This is not correct as the capital letter should be kept the way it is given.
                tokens_dict = self.to_pieces(token)
                # Transliterate words
                for token_key in tokens_dict:
                    if len(tokens_dict[token_key]):
                        word = tokens_dict[token_key]
                        if self.mode == "arabic_to_latin":
                            # w/y detection based on the priority in "word"
                            for char in word:
                                if char in self.target_char:
                                    word = self.uw_iy_Detector(word, char)  
                            if word[0] == self.hemze and word[1] in self.arabic_vowels:
                                word = word[1:]
                            word = list(word)
                            for char_index in range(len(word)):
                                word[char_index] = self.arabic_to_latin(word[char_index])
                            word = "".join(word)
                            word = self.bizroke_finder(word)
                        elif self.mode == "latin_to_arabic":
                            if len(word):
                                word = list(word)
                                for char_index in range(len(word)):
                                    word[char_index] = self.latin_to_arabic(word[char_index])
                                if word[0] in self.arabic_vowels or word[0].lower() == self.bizroke:
                                    word.insert(0, self.hemze)
                                word = "".join(word).replace("û", "وو").replace(self.bizroke.lower(), "").replace(self.bizroke.upper(), "")
                            # else:
                            #     return self.UNKNOWN

                        trans_token = trans_token + word
            
                transliterated_line.append(trans_token)
            transliterated_text.append(" ".join(transliterated_line).replace(u" w ", u" û "))

        # standardize the output
        # replace UNKOWN by the user's choice
        if self.user_UNKNOWN != self.UNKNOWN:
            return "\n".join(transliterated_text).replace(self.UNKNOWN, self.user_UNKNOWN)
        else:
            return "\n".join(transliterated_text)

    def preprocessor(self, word):
        """Preprocessing by normalizing text encoding and removing embedding characters"""
        # replace this by the normalization part
        word = list(word.replace('\u202b', "").replace('\u202c', "").replace('\u202a', "").replace(u"وو", "û").replace("\u200c", "").replace("ـ", ""))
        # for char_index in range(len(word)):
        #     if(word[char_index] in self.tricky_characters.keys()):
        #         word[char_index] = self.tricky_characters[word[char_index]]
        return "".join(word)

    def uw_iy_Detector(self, word, target_char):
        """Detection of "و" and "ی" in the Arabic-based script"""
        word = list(word)
        if target_char == "و":
            dic_index = 1
        else:
            dic_index = 0
        
        if word == target_char:
            word = self.uw_iy_forms["target_char_cons"][dic_index]
        else:
            for index in range(len(word)):
                if word[index] == self.hemze and word[index+1] == target_char:
                    word[index+1] = self.uw_iy_forms["target_char_vowel"][dic_index]
                    index += 1
                else:
                    if word[index] == target_char:
                        if index == 0:
                            word[index] = self.uw_iy_forms["target_char_cons"][dic_index]
                        else:
                            if word[index-1] in self.arabic_vowels:
                                word[index] = self.uw_iy_forms["target_char_cons"][dic_index]
                            else:
                                if index+1 < len(word):
                                    if word[index+1] in self.arabic_vowels:
                                        word[index] = self.uw_iy_forms["target_char_cons"][dic_index]
                                    else:
                                        word[index] = self.uw_iy_forms["target_char_vowel"][dic_index]
                                else:
                                    word[index] = self.uw_iy_forms["target_char_vowel"][dic_index]

        word = "".join(word).replace(self.hemze+self.uw_iy_forms["target_char_vowel"][dic_index], self.uw_iy_forms["target_char_vowel"][dic_index])
        return word

    def syllable_detector(self, word):
        """Detection of the syllable based on the given pattern. May be used for transcription applications."""
        syllable_templates = ["V", "VC", "VCC", "CV", "CVC", "CVCCC"]
        CV_converted_list = ""       
        for char in word:
            if char in self.latin_vowels:
                CV_converted_list += "V"
            else:
                CV_converted_list += "C"
        
        syllables = list()
        for i in range(1, len(CV_converted_list)):
            syllable_templates_permutated = [p for p in itertools.product(syllable_templates, repeat=i)]
            for syl in syllable_templates_permutated:
                if len("".join(syl)) == len(CV_converted_list):
                    if CV_converted_list == "".join(syl) and "VV" not in "".join(syl):
                        syllables.append(syl)
        return syllables

    def bizroke_finder(self, word):
        """Detection of the "i" character in the Arabic-based script. Incomplete version."""
        word = list(word)
        if len(word) > 2 and word[0] in self.latin_cons and word[1] in self.latin_cons and word[1] != "w" and word[1] != "y":
            word.insert(1, "i")
        return "".join(word)

    def arabic_to_latin(self, char):
        """Mapping Arabic-based characters to the Latin-based equivalents"""
        if char != "":
            if char in list(self.characters_mapping.values()):
                return list(self.characters_mapping.keys())[list(self.characters_mapping.values()).index(char)]
            elif char in self.punctuation_mapping:
                return self.punctuation_mapping[char]
            elif char in self.punctuation_mapping:
                return self.punctuation_mapping[char]
        return char

    def latin_to_arabic(self, char):
        """Mapping Latin-based characters to the Arabic-based equivalents"""
        # check if the character is in upper case
        mapped_char = ""
        
        if char.lower() != "":
            if char.lower() in self.wy_mappings.keys():
                mapped_char = self.wy_mappings[char.lower()]
            elif char.lower() in self.characters_mapping.keys():
                mapped_char = self.characters_mapping[char.lower()]
            elif char.lower() in self.punctuation_mapping:
                mapped_char = self.punctuation_mapping[char.lower()]
            # elif char.lower() in self.digits_mapping.values():
            #     mapped_char = self.digits_mapping.keys()[self.digits_mapping.values().index(char.lower())]
        
        if len(mapped_char):
            if char.isupper():
                return mapped_char.upper()
            return mapped_char
        else:
            return char


# Known bugs:
# لە ئیسپانیان ژنان لەدژی ‘patriarkavirus’ ڕێپێوانیان ئەنجامدا
# code mixed should not result an error!
# added this to the test cases. Moreover, you should fix the standardizing step after the task and preprocessing before the task
# Saza'm ne kevnar e
# Ez hîne yê bêzar im
# dwayîn zanyarî lebarey vayrosî koronay nwê (���î�-19)
# not able to transliterate words starting with capital letters : Berîtanya



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Configuration class of the Kurdish Language Processing Toolkit (KLPT).

Raises:
    ValueError: Unknown dialect
    ValueError: Unknown script
    ValueError: Unknown numeral

Returns:
    obj: an object to be passed to other packages
"""
import sys
sys.path.append('../klpt')
import json
import klpt

class Configuration:
    def __init__(self, config_dict):#dialect, script, numeral="Latin", target_script=None, unknown="�"):
        """

        Args:
            dialect (str): the name of the dialect or its ISO 639-3 code
            script (str): the name of the script
            numeral (str): the type of the numeral

        """

        with open(klpt.get_data("data/default-options.json"), encoding = "utf-8") as options_file:
            self.options = json.load(options_file)
        
        self.unknown = None

        if "script" in config_dict:
            self.validate_script(config_dict["script"])
        else:
            self.script = None

        if "dialect" in config_dict:
            self.validate_dialect(config_dict["dialect"])
        else:
            self.dialect = None

        if "numeral" in config_dict:
            self.validate_numeral(config_dict["numeral"])
        else:
            self.numeral = None
    
        if "target_script" in config_dict:
            self.validate_target_script(config_dict["target_script"])
        else:
            self.target_script = None

        if "unknown" in config_dict:
            self.validate_unknown(config_dict["unknown"])
        else:
            self.user_UNKNOWN = "�"

    def normalize_arguments(self, argument):
        return argument.lower().capitalize()

    def validate_script(self, script):
        if self.normalize_arguments(script) not in self.options["scripts"]:
            raise ValueError(f'Unknown script. Available options: {self.options["scripts"]}')
        else:
            self.script = self.normalize_arguments(script)

    def validate_dialect(self, dialect):
        if self.normalize_arguments(dialect) in self.options["dialects"]:
            self.dialect = self.normalize_arguments(dialect)
        elif self.normalize_arguments(dialect) in list(self.options["dialects"].values()):
            self.dialect = dict(zip(self.options["dialects"].values(), self.options["dialects"].keys()))[self.normalize_arguments(dialect)]
        else:
            raise ValueError(f'Unknown dialect. Available options: {self.options["dialects"]}')

    def validate_numeral(self, numeral):
        if self.normalize_arguments(numeral) not in self.options["numerals"]:
            raise ValueError(f'Unknown numeral. Available options: {self.options["numerals"]}')
        else:
            self.numeral = self.normalize_arguments(numeral).lower().capitalize()

    def validate_target_script(self, target_script):
        # complete it; if source and target scripts are identical, then return the input without processing
        if target_script == "Latin":
            self.target_script = self.normalize_arguments(target_script)
            self.mode = "arabic_to_latin"
        elif target_script == "Arabic":
            self.target_script = self.normalize_arguments(target_script)
            self.mode = "latin_to_arabic"
        else:
            raise ValueError(f'Unknown transliteration option. Available options: {self.options["transliterator"]}')

    def validate_unknown(self, unknown):
        if len(unknown):
            self.user_UNKNOWN = unknown
        else:
            raise ValueError(f'Unknown unknown tag. Select a non-empty token (e.g. <UNK>.')


# To Do:
# User should be able to update the configuration at any point in such a way that the class that uses that instance also get updated


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
sys.path.append('../klpt')
import os

__url__ = "https://sinaahmadi.github.io/klpt"

# Maintainer, contributors, etc.
__maintainer__ = "Sina Ahmadi"
__maintainer_email__ = "ahmadi.sina@outlook.com"
__author__ = __maintainer__
__author_email__ = __maintainer_email__

_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_data(path):
    return os.path.join(_ROOT, '', path)

data_directory = {
    "tokenize": {
        "Sorani": {
            "Arabic": get_data("data/lexicon_ckb_arab.json"),
            "Latin": get_data("data/lexicon_ckb_latn.json")
            },
        "Kurmanji": {
            "Latin": get_data("data/lexicon_kmr_latn.json")
        }
    },
    "morphemes": {
        "Sorani": get_data("data/ckb-morphemes.json"),
        "Kurmanji": get_data("data/kmr-morphemes.json")
    },
    "analyser": {
        "Sorani": get_data("data/ckb-analyser.att"),
        "Kurmanji": get_data("data/kmr-analyser.att")
    },
    "stopwords": get_data("data/stopwords.json")
}


#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Utility functions used in the the Kurdish Language Processing Toolkit (KLPT).

    * created: 2020/12/07 10:12:06
    * author: Sina Ahmadi

"""

import sys
sys.path.append('../klpt')

def extract_prefix_suffix(word_form, base):
    """Given a substring, find the preceding and succeeding characters as prefix and suffix, respectively

    Args:
        word_form ([str]): [a word]
        base ([str]): [the stem or base of the word]

    Returns:
        [list]: [a list containing three elements in str showing the preceding and succeeding characters before the stem or the base]
    """
    if word_form == base:
        return '', word_form, ''
    elif len(word_form) > len(base):
        for i in range(len(word_form)):
            if word_form[i: i + len(base)] == base:
                return word_form[0: i], base, word_form[i + len(base):]
    
    return '', word_form, ''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Preprocessing module of the Kurdish Language Processing Toolkit (KLPT).

    Normalizing scripts and orthographies by using writing conventions based on dialects and scripts.
    The goal is not to correct the orthography but to normalize the text in terms of the encoding and common writing rules.

        * created: 2020/05/10 02:10:54
        * author: Sina Ahmadi
    
"""

import json
import re
import sys
from .configuration import Configuration
sys.path.append('../klpt')
import klpt

class Preprocess:
    """
    This module deals with normalizing scripts and orthographies by using writing conventions based on dialects and scripts. The goal is not to correct the orthography but to normalize the text in terms of the encoding and common writing rules. The input encoding should be in UTF-8 only. To this end, three functions are provided as follows:

    - `normalize`: deals with different encodings and unifies characters based on dialects and scripts
    - `standardize`: given a normalized text, it returns standardized text based on the Kurdish orthographies following recommendations for [Kurmanji](https://books.google.ie/books?id=Z7lDnwEACAAJ) and [Sorani](http://yageyziman.com/Renusi_Kurdi.htm)
    - `unify_numerals`: conversion of the various types of numerals used in Kurdish texts
    - `preprocess`: one single function for normalization, standardization and unification of numerals

    In addition, it is possible to remove stopwords using the `stopwords` variable. It is better to remove stopwords after the tokenization task.

    It is recommended that the output of this module be used as the input of subsequent tasks in an NLP pipeline.
    
    Example:

    ```python
    >>> from klpt.preprocess import Preprocess

    >>> preprocessor_ckb = Preprocess("Sorani", "Arabic", numeral="Latin")
    >>> preprocessor_ckb.normalize("لە ســـاڵەکانی ١٩٥٠دا")
    'لە ساڵەکانی 1950دا'
    >>> preprocessor_ckb.standardize("راستە لەو ووڵاتەدا")
    'ڕاستە لەو وڵاتەدا'
    >>> preprocessor_ckb.unify_numerals("٢٠٢٠")
    '2020'
    >>> preprocessor_ckb.preprocess("راستە لە ووڵاتەی ٢٣هەمدا")
    'ڕاستە لە وڵاتەی 23هەمدا'

    >>> preprocessor_kmr = Preprocess("Kurmanji", "Latin")
    >>> preprocessor_kmr.standardize("di sala 2018-an")
    'di sala 2018an'
    >>> preprocessor_kmr.standardize("hêviya")
    'hêvîya'
    >>> preprocessor_kmr.stopwords[:10]
    ['a', 'an', 'bareya', 'bareyê', 'barên', 'basa', 'be', 'belê', 'ber', 'bereya']
    ```

    The preprocessing rules are provided at [`data/preprocess_map.json`](https://github.com/sinaahmadi/klpt/blob/master/klpt/data/preprocess_map.json).
    """

    def __init__(self, dialect, script, numeral="Latin"):
        """
        Initialization of the Preprocess class

        Arguments:
            dialect (str): the name of the dialect or its ISO 639-3  code
            script (str): the name of the script
            numeral (str): the type of the numeral
        
        """
        with open(klpt.get_data("data/preprocess_map.json"), encoding = "utf-8") as preprocess_file:
            self.preprocess_map = json.load(preprocess_file)

        configuration = Configuration({"dialect": dialect, "script": script, "numeral": numeral})
        self.dialect = configuration.dialect
        self.script = configuration.script
        self.numeral = configuration.numeral
        # self.preprocess_map = config.preprocess_map

        with open(klpt.data_directory["stopwords"], "r", encoding = "utf-8") as f:
            self.stopwords = json.load(f)[dialect][script]

    def standardize(self, text):
        """
        Method of standardization of Kurdish orthographies

        Given a normalized text, it returns standardized text based on the Kurdish orthographies.

        - Sorani-Arabic:
            - replace alveolar flap ر (/ɾ/) at the begging of the word by the alveolar trill ڕ (/r/)
            - replace double rr and ll with ř and ł respectively

        - Kurmanji-Latin:
            - replace "-an" or "'an" in dates and numerals ("di sala 2018'an" and "di sala 2018-an" -> "di sala 2018an")

        Open issues:
            - replace " وە " by  " و "? But this is not always possible, "min bo we" (ریزگـرتنا من بو وە  نە ئە وە ئــە ز)
            - "pirtükê": "pirtûkê"?
            - Should [ı (LATIN SMALL LETTER DOTLESS I](https://www.compart.com/en/unicode/U+0131) be replaced by i?

        Arguments:
            text (str): a string

        Returns:
            str: standardized text

        """
        temp_text = " " + self.unify_numerals(text) + " "

        for standardization_type in [self.dialect]:
            for rep in self.preprocess_map["standardizer"][standardization_type][self.script]:
                rep_tar = self.preprocess_map["standardizer"][standardization_type][self.script][rep]
                temp_text = re.sub(rf"{rep}", rf"{rep_tar}", temp_text, flags=re.I)

        return temp_text.strip()

    def normalize(self, text):
        """
        Text normalization

        This function deals with different encodings and unifies characters based on dialects and scripts as follows:
        
        - Sorani-Arabic:
        
            - replace frequent Arabic characters with their equivalent Kurdish ones, e.g. "ي" by "ی" and "ك" by "ک"
            - replace "ه" followed by zero-width non-joiner (ZWNJ, U+200C) with "ە" where ZWNJ is removed ("ره‌زبه‌ر" is converted to "رەزبەر"). ZWNJ in HTML is also taken into account.
            - replace "هـ" with "ھ" (U+06BE, ARABIC LETTER HEH DOACHASHMEE)
            - remove Kashida "ـ"
            - "ھ" in the middle of a word is replaced by ه (U+0647)
            - replace different types of y, such as 'ARABIC LETTER ALEF MAKSURA' (U+0649)

        It should be noted that the order of the replacements is important. Check out provided files for further details and test cases.

        Arguments:
            text (str): a string

        Returns:
            str: normalized text

         """
        temp_text = " " + self.unify_numerals(text) + " "

        for normalization_type in ["universal", self.dialect]:
            for rep in self.preprocess_map["normalizer"][normalization_type][self.script]:
                rep_tar = self.preprocess_map["normalizer"][normalization_type][self.script][rep]
                temp_text = re.sub(rf"{rep}", rf"{rep_tar}", temp_text, flags=re.I)

        return temp_text.strip()

    def unify_numerals(self, text):
        """
        Convert numerals to the desired one

        There are three types of numerals:
        - Arabic [١٢٣٤٥٦٧٨٩٠]
        - Farsi [۱۲۳۴۵۶۷۸۹۰]
        - Latin [1234567890]

        Arguments:
            text (str): a string

        Returns:
            str: text with unified numerals

        """
        for i, j in self.preprocess_map["normalizer"]["universal"]["numerals"][self.numeral].items():
            text = text.replace(i, j)
        return text

    def preprocess(self, text):
        """
        One single function for normalization, standardization and unification of numerals

        Arguments:
            text (str): a string

        Returns:
            str: preprocessed text
        """
        return self.unify_numerals(self.standardize(self.normalize(text)))



from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding = "utf-8") as f:
    long_description = f.read()

with open('requirements.txt', encoding = "utf-8") as f:
    required = [req.strip() for req in f.read().splitlines() if req.strip()]

setup(
    name="klpt",
    version="0.1.6",
    description="Kurdish Language Processing Toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Sina Ahmadi",
    author_email="ahmadi.sina@outlook.com",
    url="https://github.com/sinaahmadi/klpt",
    packages=find_packages(exclude=["tests", "cinder"]),
    license="CC BY-SA 4.0",
    install_requires=required,
    include_package_data=True,
    python_requires=">=3.6"
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" The stemming module of the Kurdish Language Processing Toolkit (KLPT).

Based on the Hunspell implementation of Sorani Kurdish by Sina Ahmadi (2020)
    * created: 2020/07/12 02:58:26
    * author: Sina Ahmadi

"""

import sys
import json
from hunspell import Hunspell
from klpt.att_analyze import Analysis
from klpt.configuration import Configuration
sys.path.append('../klpt')
import klpt
from klpt import utility

class Stem:
    """

    The Stem module deals with various tasks, mainly through the following functions:
        - `check_spelling`: spell error detection
        - `correct_spelling`: spell error correction
        - `analyze`: morphological analysis
        - `stem`: stemming, e.g. "بڕاوە" → "بڕ"
        - `lemmatize`: lemmatization, e.g. "بردمنەوە" → "بردن"

    It is recommended that this module be used on tokens using the tokenization module. 
    Please note that only Sorani is supported in this version in this module. The module is based on the [Kurdish Hunspell project](https://github.com/sinaahmadi/KurdishHunspell).

    Regarding stemming, the following procedure is followed:
        - for tokens of a single word, as "kirin" (to do), the stem of the token is returned.
        - for compound forms and multi-word expressions, the stem of the noun, adjective or adverb are taken into account. For instance, in the light verbal constructions such as "bar kirin" (to load), the stem of the nominal component "bar" is returned. In other cases, the stem of that part of the MWE token is returned that is semantically more important, as in "دەست تێ وەردان" (dest-tê-werdan) where the stem of "dest" is returned.

    Example:
    ```python
    >>> from klpt.stem import Stem
    >>> stemmer = Stem("Sorani", "Arabic")
    >>> stemmer.check_spelling("سوتاندبووت")
    False
    >>> stemmer.correct_spelling("سوتاندبووت")
    (False, ['ستاندبووت', 'سووتاندبووت', 'سووڕاندبووت', 'ڕووتاندبووت', 'فەوتاندبووت', 'بووژاندبووت'])
    >>> stemmer.analyze("دیتبامن")
    [{'pos': ['verb'], 'description': 'past_stem_transitive_active', 'stem': 'دی', 'lemma': ['دیتن'], 'base': 'دیت', 'prefixes': '', 'suffixes': 'بامن'}]
    >>> stemmer.stem("دەچینەوە")
    ['چ']
    >>> stemmer.stem("گورەکە", mark_unknown=True)
    ['_گور_']
    >>> stemmer.lemmatize("گوڵەکانم")
    ['گوڵ', 'گوڵە']
    
    >>> stemmer = Stem("Kurmanji", "Latin")
    >>> stemmer.analyze("dibêjim")
    [{'base': 'gotin', 'description': 'vblex_tv_pri_p1_sg', 'pos': '', 'terminal_suffix': '', 'formation': ''}]
    ```

    """
    # to do: make the following function work for Kurmanji Latin also following the same algorithm
    def __init__(self, dialect, script):

        self.dialect = dialect
        self.script = script 

        self.hunspell_flags = {"po": "pos", "is": "description", "ds": "formation", "st": "stem", "lem": "lemma"}
        
        if self.dialect == "Sorani" and self.script == "Arabic":
            self.huns = Hunspell("ckb-Arab", hunspell_data_dir=klpt.get_data("data/"))
            with open(klpt.data_directory["morphemes"][self.dialect], "r", encoding = "utf-8") as f_morphemes:
                self.light_verbs = json.load(f_morphemes)["Morphemes"]["light_verbs"][self.script]
        elif self.dialect == "Kurmanji" and self.script == "Latin":
            self.huns = Hunspell("kmr-Latn", hunspell_data_dir=klpt.get_data("data/"))
            with open(klpt.data_directory["morphemes"][self.dialect], "r", encoding = "utf-8") as f_morphemes:
                self.light_verbs = json.load(f_morphemes)["Morphemes"]["light_verbs"][self.script]
        else:
            raise Exception("Sorry, only Sorani dialect in the Arabic script and Kurmanji in the Latin script is supported now. Stay tuned for other dialects and scripts!")
            
        with open(klpt.data_directory["morphemes"][self.dialect], "r", encoding = "utf-8") as f_morphemes:
            self.morphemes = json.load(f_morphemes)["Morphemes"]["Concatenated"][self.script]

    def stem(self, word, mark_unknown=False):
        """A function for stemming a single word

        Args:
            word (str): input word to be spell-checked
            mark_unknown (False): if the given word is unknown in the tagged lexicon, KLPT stems is following rules. Such stems can be marked with "_" if this variable set to True

        Raises:
            TypeError: only string as input

        Returns:
            list: list of stem(s)
        """
        if not isinstance(word, str):
            raise TypeError("Not supported yet.")
        elif (self.dialect == "Sorani" and self.script == "Arabic") or (self.dialect == "Kurmanji" and self.script == "Latin"):
            stems = list(set([self.clean_stem(i) for i in self.huns.stem(word)]))
            if len(stems):
                return stems
            else:
                # not detected by Hunspell or the word doesn't exist in the tagged lexicon
                for verb in self.light_verbs:
                    if word.endswith(verb) and len(word.rpartition(verb)[0]):
                        stems = list(set([self.clean_stem(i) for i in self.huns.stem(word.rpartition(verb)[0].strip())]))
                        if len(stems):
                            # the word is a compound form with a light verb. The other part can be stemmed by Hunspell
                            return stems
                        else:
                            # the word is a compound form with a light verb but the other part cannot be stemmed by Hunspell
                            word = word.rpartition(verb)[0].strip()
                
                # the other part of the word or the whole word cannot be stemmed by Hunspell
                # so, find the stem following morphological rules by checking if removing possible prefixes and suffixes would help finding the stem.
                # Note: even though the same morphemes used in the tokenization system are used in the rules here, there is a delicate difference.
                #    In the tokenization system, the trimming is done in such a way that shorter morphemes are first checked for suffixes (suffixes in the json file is sorted by length) and longer prefixes are trimmed first.
                #    For the stemmer, however, we do differently by first checking the longer morphemes then shorter ones (for both prefixes and suffixes). 
                #    This is due to the different purposes of the two tasks. Therefore, the list of the morphemes is to be reversed for suffixes (not prefixes). 
                # In order not to modify the json files, the `reversed` function is used for suffixes.
                
                for preposition in self.morphemes["prefixes"]:
                    if word.startswith(preposition) and len(word.split(preposition, 1)) > 1:
                        if len(list(set([self.clean_stem(i) for i in self.huns.stem(word.split(preposition, 1)[1])]))):
                            stems = list(set([self.clean_stem(i) for i in self.huns.stem(word.split(preposition, 1)[1])]))
                            if mark_unknown:
                                return ["_" + i + "_" for i in stems]
                        else:
                            word = word.split(preposition, 1)[1]
                            break
                
                for postposition in reversed(list(self.morphemes["suffixes"])):
                    if word.endswith(postposition) and len(word.rpartition(postposition)[0]):
                        if len(list(set([self.clean_stem(i) for i in self.huns.stem(word.rpartition(postposition)[0])]))):
                            stems = list(set([self.clean_stem(i) for i in self.huns.stem(word.rpartition(postposition)[0])]))
                            if mark_unknown:
                                return ["_" + i + "_" for i in stems]
                        else:
                            word = word.rpartition(postposition)[0]
                            break
                
                # not possible to stem the word using the tagged lexicon or the rule-based approach. Return the word as it is.
                if mark_unknown:
                    return ["_" + word + "_"]
                else:
                    return [word]

    def lemmatize(self, word):
        """A function for lemmatization of words

        Args:
            word ([str]): [given a word, return its lemma form, i.e. dictionary entry form]
        
        Raises:
            TypeError: only string as input

        Returns:
            list: list of lemma(s)
        """
        if not isinstance(word, str):
            raise TypeError("Not supported yet.")
        elif (self.dialect == "Sorani" and self.script == "Arabic") or (self.dialect == "Kurmanji" and self.script == "Latin"):
            word_analysis = self.analyze(word)
            return list(set([item for sublist in word_analysis for item in sublist["lemma"] if item != '']))

    def clean_stem(self, word):
        """Remove extra characters in the stem
        The following issue was observed when stemming with Hunspell (version 2.0.2) where
        the retrieved stem of a verb is accompanied by the flag of the word, which is an unwanted extra character.
        Possible flags are T, V and I. :lf should also be taken into account.

        Args:
            word ([str]): [stem]
        """
        for char in ["V", "I", "T"]:
            word = word.replace(char, "")
        return word.replace(":lf", "")

    def check_spelling(self, word):
        """Check spelling of a word

        Args:
            word (str): input word to be spell-checked

        Raises:
            TypeError: only string as input

        Returns:
            bool: True if the spelling is correct, False if the spelling is incorrect
        """
        if not isinstance(word, str):
            raise TypeError("Not supported yet.")
        elif (self.dialect == "Sorani" and self.script == "Arabic") or (self.dialect == "Kurmanji" and self.script == "Latin"):
            return self.huns.spell(word)

    def correct_spelling(self, word):
        """
        Correct spelling errors if the input word is incorrect. It returns a tuple where the first element indicates the correctness of the word (True if correct, False if incorrect).
            If the input word is incorrect, suggestions are provided in a list as the second element of the tuple, as (False, []).
            If no suggestion is available, the list is returned empty as (True, []).

        Args:
            word (str): input word to be spell-checked

        Raises:
            TypeError: only string as input

        Returns:
            tuple (boolean, list)

        """
        if not isinstance(word, str):
            raise TypeError("Not supported yet.")
        elif (self.dialect == "Sorani" and self.script == "Arabic") or (self.dialect == "Kurmanji" and self.script == "Latin"):
            if self.check_spelling(word):
                return (True, [])
            return (False, list(self.huns.suggest(word)))

    def analyze(self, word_form):
        """
        Morphological analysis of a given word.
        
        It returns morphological analyses. The morphological analysis is returned as a dictionary as follows:
        
        - "pos": the part-of-speech of the word-form according to [the Universal Dependency tag set](https://universaldependencies.org/u/pos/index.html). The gender for nouns in Kurmanji is also provided after the "noun" tag.
        - "description": is flag
        - "prefixes": anything appearing before the base
        - "suffixes": anything appearing after the base
        - "st": the stem of the word
        - "lem": the lemma of the word
        - "formation": if ds flag is set, its value is assigned to description and the value of formation is set to derivational. Although the majority of our morphological rules cover inflectional forms, it is not accurate to say all of them are inflectional. Therefore, we only set this value to derivational wherever we are sure.
        - "base": `ts` flag. The definition of terminal suffix is a bit tricky in Hunspell. According to [the Hunspell documentation](http://manpages.ubuntu.com/manpages/trusty/en/man4/hunspell.4.html), "Terminal suffix fields are inflectional suffix fields "removed" by additional (not terminal) suffixes". In other words, the ts flag in Hunspell represents whatever is left after stripping all affixes. Therefore, it is the morphological base.

        As for the word "دیتبامن" (that I have seen them), the morphological analysis would look like this: [{'pos': ['verb'], 'description': 'past_stem_transitive_active', 'stem': 'دی', 'lemma': ['دیتن'], 'base': 'دیت', 'prefixes': '', 'suffixes': 'بامن'}]
        If the input cannot be analyzed morphologically, an empty list is returned.

        More details regarding Sorani Kurdish morphological analysis can be found at [https://github.com/sinaahmadi/KurdishHunspell](https://github.com/sinaahmadi/KurdishHunspell).

        Nota bene:
        In the previous versions of KLPT, the `stem` module for Kurmanji was relied on the [Apertium project](https://github.com/apertium/apertium-kmr). Now, that is fully replaced by the Kurmanji implementation of [Kurdish Hunspell](https://github.com/sinaahmadi/KurdishHunspell). 

        Args:
            word_form (str): a single word-form

        Raises:
            TypeError: only string as input

        Returns:
            (list(dict)): a list of all possible morphological analyses according to the defined morphological rules
            
        """
        if not isinstance(word_form, str):
            raise TypeError("Only a word (str) is allowed.")
        else:
            word_analysis = list()
            # Given the morphological analysis of a word-form with Hunspell flags, extract relevant information and return a dictionary
            # print(self.huns.analyze(word_form))
            for analysis in list(self.huns.analyze(word_form)):
                analysis_dict = dict()
                for item in analysis.split():
                    if ":" not in item:
                        continue
                    if item.split(":")[1] == "ts":
                        # ts flag exceptionally appears after the value as value:key in the Hunspell output
                        # anything except the terminal_suffix (ts) is considered to be the base
                        analysis_dict["base"] = item.split(":")[0]
                        affixes = utility.extract_prefix_suffix(word_form, item.split(":")[0])
                        analysis_dict["prefixes"] = affixes[0]
                        analysis_dict["suffixes"] = affixes[2]
                        
                    elif item.split(":")[0] in self.hunspell_flags.keys():
                        # assign the key:value pairs from the Hunspell string output to the dictionary output of the current function
                        if item.split(":")[0] == "ds":
                            # for ds flag, add derivation as the formation type, otherwise inflection
                            analysis_dict[self.hunspell_flags[item.split(":")[0]]] = "derivational"
                            analysis_dict[self.hunspell_flags["is"]] = item.split(":")[1]

                        elif item.split(":")[0] == "st":
                            # for st flag, stem should be cleaned first
                            analysis_dict[self.hunspell_flags[item.split(":")[0]]] = self.clean_stem(item.split(":")[1])

                        else:
                            # remove I, T or V using clean_stem()
                            analysis_dict[self.hunspell_flags[item.split(":")[0]]] = self.clean_stem(item.split(":")[1])
                            
                # convert lemma and pos to a list and split based on _ when there is more than one output, e.g. more than one lemma for a given word
                if "lemma" in analysis_dict:
                    analysis_dict["lemma"] = analysis_dict["lemma"].split("_")
                else:
                    analysis_dict["lemma"] = [""]
                
                if "pos" in analysis_dict:
                    analysis_dict["pos"] = analysis_dict["pos"].split("_")
                else:
                    analysis_dict["pos"] = [""]
                
                # for nouns, base is lemma
                if len(analysis_dict["pos"]) and analysis_dict["pos"] != ["verb"]:
                    analysis_dict["lemma"] = [analysis_dict["base"]]

                word_analysis.append(analysis_dict)

        return word_analysis


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""The tokenization module of the Kurdish Language Processing Toolkit (KLPT)
    Based on the Kurdish Tokenization project: https://github.com/sinaahmadi/KurdishTokenization

    
    Regarding multi-word expressions, we assume that each part of a compound form is separated with a dash (-). Therefore, hyphens are not considered a punctuation mark in the tokenization process.

    * 2020/07/23 20:01:05
    * Sina Ahmadi

"""
import json
import sys
import re
sys.path.append('../klpt')
from klpt.configuration import Configuration
from klpt.preprocess import Preprocess
import klpt

class Tokenize:
    """

    This module focuses on the tokenization of both Kurmanji and Sorani dialects of Kurdish with the following functions:

    - `word_tokenize`: tokenization of texts into tokens (both [multi-word expressions](https://aclweb.org/aclwiki/Multiword_Expressions) and single-word tokens).
    - `mwe_tokenize`: tokenization of texts by only taking compound forms into account
    - `sent_tokenize`: tokenization of texts into sentences

    The module is based on the [Kurdish tokenization project](https://github.com/sinaahmadi/KurdishTokenization).

    Example:
    
    ```python
    >>> from klpt.tokenize import Tokenize

    >>> tokenizer = Tokenize("Kurmanji", "Latin")
    >>> tokenizer.word_tokenize("ji bo fortê xwe avêtin")
    ['▁ji▁', 'bo', '▁▁fortê‒xwe‒avêtin▁▁']
    >>> tokenizer.mwe_tokenize("bi serokê hukûmeta herêma Kurdistanê Prof. Salih re saz kir.")
    'bi serokê hukûmeta herêma Kurdistanê Prof . Salih re saz kir .'

    >>> tokenizer_ckb = Tokenize("Sorani", "Arabic")
    >>> tokenizer_ckb.word_tokenize("بە هەموو هەمووانەوە ڕێک کەوتن")
    ['▁بە▁', '▁هەموو▁', 'هەمووانەوە', '▁▁ڕێک‒کەوتن▁▁']
    ```

    """
    def __init__(self, dialect, script, numeral="Latin", separator='▁'):

        # validate parameters
        with open(klpt.get_data("data/tokenize.json"), encoding = "utf-8") as tokenize_file:
            self.tokenize_map = json.load(tokenize_file)

        with open(klpt.get_data("data/preprocess_map.json"), encoding = "utf-8") as preprocess_file:
            self.preprocess_map = json.load(preprocess_file)

        # sentence tokenizer variables
        self.dialect, self.script = dialect, script
        self.alphabets = "([%s])"%"".join(self.tokenize_map["sent_tokenize"][self.dialect][self.script]["alphabet"])
        self.prefixes = self.tokenize_map["sent_tokenize"][self.dialect][self.script]["prefixes"]
        self.suffixes = self.tokenize_map["sent_tokenize"][self.dialect][self.script]["suffixes"]
        self.starters = self.tokenize_map["sent_tokenize"][self.dialect][self.script]["starters"]
        self.websites = self.tokenize_map["sent_tokenize"]["universal"]["websites"]
        self.acronyms = self.tokenize_map["sent_tokenize"][self.dialect][self.script]["acronyms"]
        self.digits = "([%s])"%"".join(list(set(list(self.preprocess_map["normalizer"]["universal"]["numerals"][numeral].values()))))

        # load lexicons
        with open(klpt.data_directory["tokenize"][self.dialect][self.script], "r", encoding = "utf-8") as f_lexicon:
            self.lexicon = json.load(f_lexicon)["Lexicon"]
         
        self.mwe_lexicon = {lemma: form for lemma, form in self.lexicon.items() if "-" in lemma}

        with open(klpt.data_directory["morphemes"][self.dialect], "r", encoding = "utf-8") as f_morphemes:
            self.morphemes = json.load(f_morphemes)["Morphemes"]["Concatenated"][self.script]
        
    
    def mwe_tokenize(self, sentence, separator="▁▁", in_separator="‒", punct_marked=False, keep_form=False):
        """
        Multi-word expression tokenization

        Args:
            sentence (str): sentence to be split by multi-word expressions
            separator (str): a specific token to specify a multi-word expression. By default two ▁ (▁▁) are used for this purpose.
            in_separator (str): a specific token to specify the composing parts of a multi-word expression. By default a dash - is used for this purpose.
            keep_form (boolean): if set to True, the original form of the multi-word expression is returned the same way provided in the input. On the other hand, if set to False, the lemma form is used where the parts are delimited by a dash ‒, as in "dab‒û‒nerît"

        Returns:
            str: sentence containing d multi-word expressions using the separator

        """
        sentence = " " + sentence + " "

        if not punct_marked:
            # find punctuation marks and add a space around
            for punct in self.tokenize_map["word_tokenize"][self.dialect][self.script]["punctuation"]:
                if punct in sentence:
                    sentence = sentence.replace(punct, " " + punct + " ")

        # look for compound words and delimit them by double separator
        for compound_lemma in self.mwe_lexicon:
            compound_lemma_context = " " + compound_lemma + " "
            if compound_lemma_context in sentence:
                if keep_form:
                    sentence = sentence.replace(compound_lemma_context, " ▁▁" + compound_lemma + "▁▁ ")
                else:
                    sentence = sentence.replace(compound_lemma_context, " ▁▁" + compound_lemma.replace("-", in_separator) + "▁▁ ")                    

            # check the possible word forms available for each compound lemma in the lex files, too
            # Note: compound forms don't have any hyphen or separator in the lex files
            for compound_form in self.mwe_lexicon[compound_lemma]["token_forms"]:
                compound_form_context = " " + compound_form + " "
                if compound_form_context in sentence:
                    if keep_form:
                        sentence = sentence.replace(compound_form_context, " ▁▁" + compound_form + "▁▁ ")
                    else:
                        sentence = sentence.replace(compound_form_context, " ▁▁" + compound_lemma.replace("-", in_separator) + "▁▁ ")
                        

        # print(sentence)
        return sentence.replace("  ", " ").replace("▁▁", separator).strip()
        

    def word_tokenize(self, sentence, separator="▁", mwe_separator="▁▁", keep_form=False):
        """Word tokenizer

        Args:
            sentence (str): sentence or text to be tokenized

        Returns:
            [list]: [a list of words]
       
        """
        # find multi-word expressions in the sentence
        sentence = self.mwe_tokenize(sentence, keep_form=keep_form)

        # find punctuation marks and add a space around
        for punct in self.tokenize_map["word_tokenize"][self.dialect][self.script]["punctuation"]:
            if punct in sentence:
                sentence = sentence.replace(punct, " " + punct + " ")

        # print(sentence)
        tokens = list()
        # split the sentence by space and look for identifiable tokens
        for word in sentence.strip().split():
            if "▁▁" in word:
                # the word is previously detected as a compound word
                tokens.append(word)
            else:
                if word in self.lexicon:
                    # check if the word exists in the lexicon
                    tokens.append("▁" + word + "▁")
                else:
                    # the word is neither a lemma nor a compound
                    # morphological analysis by identifying affixes and clitics
                    token_identified = False

                    for preposition in self.morphemes["prefixes"]:
                        if word.startswith(preposition) and len(word.split(preposition, 1)) > 1:
                            if word.split(preposition, 1)[1] in self.lexicon:
                                word = "▁".join(["", self.morphemes["prefixes"][preposition], word.split(preposition, 1)[1], ""])
                                token_identified = True
                                break
                            elif self.mwe_tokenize(word.split(preposition, 1)[1], keep_form=keep_form) != word.split(preposition, 1)[1]:
                                word = "▁" + self.morphemes["prefixes"][preposition] + self.mwe_tokenize(word.split(preposition, 1)[1], keep_form=keep_form)
                                token_identified = True
                                break
                    
                    if not token_identified:
                        for postposition in self.morphemes["suffixes"]:
                            if word.endswith(postposition) and len(word.rpartition(postposition)[0]):
                                if word.rpartition(postposition)[0] in self.lexicon:
                                    word = "▁" + word.rpartition(postposition)[0] + "▁" + self.morphemes["suffixes"][postposition]
                                    break
                                elif self.mwe_tokenize(word.rpartition(postposition)[0], keep_form=keep_form) != word.rpartition(postposition)[0]:
                                    word = ("▁" + self.mwe_tokenize(word.rpartition(postposition)[0], keep_form=keep_form) + "▁" + self.morphemes["suffixes"][postposition] + "▁").replace("▁▁▁", "▁▁")
                                    break
            
                    tokens.append(word)
        # print(tokens)
        return " ".join(tokens).replace("▁▁", mwe_separator).replace("▁", separator).split()


    def sent_tokenize(self, text):
        """Sentence tokenizer

        Args:
            text ([str]): [input text to be tokenized by sentences]

        Returns:
            [list]: [a list of sentences]

        """
        text = " " + text + "  "
        text = text.replace("\n", " ")
        text = re.sub(self.prefixes, "\\1<prd>", text)
        text = re.sub(self.websites, "<prd>\\1", text)
        text = re.sub("\s" + self.alphabets + "[.] ", " \\1<prd> ", text)
        text = re.sub(self.acronyms + " " + self.starters, "\\1<stop> \\2", text)
        text = re.sub(self.alphabets + "[.]" + self.alphabets + "[.]" + self.alphabets + "[.]", "\\1<prd>\\2<prd>\\3<prd>", text)
        text = re.sub(self.alphabets + "[.]" + self.alphabets + "[.]", "\\1<prd>\\2<prd>", text)
        text = re.sub(" " + self.suffixes + "[.] " + self.starters, " \\1<stop> \\2", text)
        text = re.sub(" " + self.suffixes + "[.]", " \\1<prd>", text)
        text = re.sub(self.digits + "[.]" + self.digits, "\\1<prd>\\2", text)
        
        # for punct in self.tokenize_map[self.dialect][self.script]["compound_puncts"]:
        #     if punct in text:
        #         text = text.replace("." + punct, punct + ".")
        
        for punct in self.tokenize_map["sent_tokenize"][self.dialect][self.script]["punct_boundary"]:
            text = text.replace(punct, punct + "<stop>")
        
        text = text.replace("<prd>", ".")
        sentences = text.split("<stop>")
        sentences = [s.strip() for s in sentences if len(s.strip())]
        
        return sentences